from alpyperl.examples.cartpole_v0 import CartPoleEnv
from ray.rllib.algorithms.ppo import PPOConfig


policy = (
    PPOConfig()
    .rollouts(
        num_rollout_workers=2,
        num_envs_per_worker=2,
        ignore_worker_failures=True,
        recreate_failed_workers=True,
        num_consecutive_worker_failures_tolerance=3
    )
    .environment(
        CartPoleEnv, 
        env_config={
            'run_exported_model': True,
            'exported_model_loc': './resources/exported_models/cartpole_v0',
            'show_terminals': False,
            'verbose': False
        }
    )
    .build()
)

for _ in range(10):
    result = policy.train()

checkpoint_dir = policy.save("./resources/trained_policies/cartpole_v0")
print(f"Checkpoint saved in directory '{checkpoint_dir}'")

# Close all enviornments
policy.stop()


import math
from gymnasium import spaces
import numpy as np
from alpyperl import BaseAnyLogicEnv


class CartPoleEnv(BaseAnyLogicEnv):

    def __init__(self, env_config=None, disable_env_checking=True):

        # ---------------------------------------------------------------------
        # Action and observation spaces
        # ---------------------------------------------------------------------

        # Thresholds
        theta_threshold_radians = 12 * 2 * math.pi / 360.0
        x_threshold = 2.4
        # Observation space array thresholds
        high = np.array(
            [
                x_threshold * 2,
                np.finfo(np.float32).max,
                theta_threshold_radians * 2,
                np.finfo(np.float32).max
            ]
        )
        # Action/observation spaces
        self.action_space = spaces.Box(np.array([0, -1]), np.array([1, 0]), dtype=np.float32)
        self.observation_space = spaces.Box(-high, high, dtype=np.float32)
        
        # Initialise AnyLogic environment experiment
        super(CartPoleEnv, self).__init__(env_config)


from alpyperl.serve.rllib import launch_policy_server
from alpyperl.examples.cartpole_v2 import CartPoleEnv
from ray.rllib.algorithms.ppo import PPOConfig


# Launch server
launch_policy_server(
    policy_config=PPOConfig(),
    env=CartPoleEnv,
    trained_policy_loc='./resources/trained_policies/cartpole_v2/checkpoint_000012',
    port=3000
)


from ray.rllib.algorithms.ppo import PPOConfig
import math
from gymnasium import spaces
import numpy as np
from alpyperl import create_custom_env

if __name__ == "__main__":

    # The cartpole ACTION and OBSERVATION spaces
    # ------------------------------------------

    # Positional thresholds
    theta_threshold_radians = 12 * 2 * math.pi / 360.0
    x_threshold = 2.4
    # Create observation space array thresholds
    high = np.array(
        [
            x_threshold * 2,            # Horizontal position
            np.finfo(np.float32).max,   # Linear speed
            theta_threshold_radians * 2,# Pole angle
            np.finfo(np.float32).max    # Angular velocity
        ]
    )
    # Create Action and Observation spaces using `gymnasium.spaces`
    action_space = spaces.Discrete(2)
    observation_space = spaces.Box(-high, high, dtype=np.float32)


    # The TRAIN script
    # ------------------------------------------
    policy = (
        PPOConfig()
        .rollouts(
            num_rollout_workers=2,
            num_envs_per_worker=2,
            ignore_worker_failures=True,
            recreate_failed_workers=True,
            num_consecutive_worker_failures_tolerance=3
        )
        .environment(
            create_custom_env(action_space, observation_space), 
            env_config={
                'run_exported_model': True,
                'exported_model_loc': './resources/exported_models/cartpole_v0',
                'show_terminals': False,
                'verbose': False
            }
        )
        .build()
    )

    for _ in range(10):
        result = policy.train()

    checkpoint_dir = policy.save("./resources/trained_policies/cartpole_v0")
    print(f"Checkpoint saved in directory '{checkpoint_dir}'")

    # Close all enviornments
    policy.stop()


import logging
import gymnasium as gym
from gymnasium import spaces
from alpyperl.anylogic.model.connector import AnyLogicModelConnector
import numpy as np

def create_custom_env(action_space, observation_space, env_config: dict=None):
    """ Create a custom environment by passing an `action` and `observation`

    :param action_space: A valid action space: integer, double or an array of doubles
    :type action_space: gymnasium.spaces
    :param observation_space: A valid observation space as an array of doubles
    :type observation_space: gymnasium.spaces.Box
    :param env_config: Environment configuration which includes:

        * ``'run_exported_model'``: In case you want to run an exported version
        of the model. Otherwise it will wait for the AnyLogic model to connect.

        * ``'exported_model_loc'``: The location of the exported model folder.

        * ``'show_terminals'``: This only applies if running an exported model
        and the user wants a terminal to be launched for every model instance 
        (could be useful for debugging purposes).

        * ``'server_mode_on'``: This is for internal use only. It is used to 
        flag the AnyLogic model to not be launched when serving a trained policy.
        
        * ``'verbose'``: To be activated in case DEBUG logger wants to be 
        activated.

    :type env_config: dict

    :return: Returns a class definition of your custom environment with the 
        specified action and observation spaces
    :rtype: CustomEnv
    """
    class CustomEnv(BaseAnyLogicEnv):

        def __init__(self, env_config=None):
            # Action/observation spaces
            self.action_space = action_space
            self.observation_space = observation_space
            # Initialise AnyLogic environment experiment
            super(CustomEnv, self).__init__(env_config)

    return CustomEnv


class BaseAnyLogicEnv(gym.Env):
    """
    The python class that contains the AnyLogic model connection and is in 
    charge of retrieving the information required to be returned by `OpenAI 
    Gymnasium` functions such as `step` and `reset`.
    """

    metadata = {'render.modes': ['human']}

    def __init__(
        self,
        env_config: dict = {
            'run_exported_model': True,
            'exported_model_loc': './exported_model',
            'show_terminals': False,
            'server_mode_on': False,
            'verbose': False
        },
        disable_env_checking: bool = True
    ):
        """
        Internal AnyLogic environment wrapper constructor

        :param env_config: Environment configuration which includes:

            * ``'run_exported_model'``: In case you want to run an exported version
            of the model. Otherwise it will wait for the AnyLogic model to connect.

            * ``'exported_model_loc'``: The location of the exported model folder.

            * ``'show_terminals'``: This only applies if running an exported model
            and the user wants a terminal to be launched for every model instance 
            (could be useful for debugging purposes).

            * ``'server_mode_on'``: This is for internal use only. It is used to 
            flag the AnyLogic model to not be launched when serving a trained policy.
            
            * ``'verbose'``: To be activated in case DEBUG logger wants to be 
            activated.

        :type env_config: dict
        """
        # Initialise `env_config` to avoid problems when handling `None`
        self.env_config = env_config if env_config is not None else []

        # Initialise logger
        verbose = (
            'verbose' in self.env_config
            and self.env_config['verbose']
        )
        # Only log message from `alpyperl`
        ch = logging.StreamHandler()
        ch.addFilter(logging.Filter('alpyperl'))
        # Create logger configuration
        logging.basicConfig(
            level=logging.DEBUG if verbose else logging.INFO,
            format=f"%(asctime)s [%(name)s][%(levelname)8s] %(message)s",
            handlers=[ch],
        )
        self.logger = logging.getLogger(__name__)

        # Check if server mode is on.
        # When loading a trained policy, there's no need to launch the model
        # as it causes an overhead.
        # Yet, obervation space is required to be returned. In that case,
        # it is only necessary to return a sample.
        self.server_mode_on = (
            'server_mode_on' in self.env_config
            and self.env_config['server_mode_on']
        )
        # Launch or connect to AnyLogic model using the connector and launcher.
        if not self.server_mode_on:
            self.anylogic_connector = AnyLogicModelConnector(
                run_exported_model=(
                    self.env_config['run_exported_model'] 
                    if 'run_exported_model' in self.env_config 
                    else True
                ),
                exported_model_loc=(
                    self.env_config['exported_model_loc'] 
                    if 'exported_model_loc' in self.env_config 
                    else './exported_model'
                ),
                show_terminals=(
                    self.env_config['show_terminals'] 
                    if 'show_terminals' in self.env_config 
                    else False
                )
            )
            # The gateway is the direct interface to the AnyLogic model.
            self.anylogic_model = self.anylogic_connector.gateway

            # Initialise and prepare the model by calling `reset` method.
            self.anylogic_model.reset()
            
            self.logger.info("AnyLogic model has been initialized correctly!")

    def step(self, action):
        """`[INTERNAL]` Basic function for performing 'steps' in order for the simulation to
        move on. It requires an `action` as an input. This action can be of
        different types (including an array of values).
        """

        # Run fast simulation until next action is required (which will be
        # controlled and requested from the AnyLogic model)
        if not self.server_mode_on:
            # Parse action to a type that can be consumed by AnyLogic model.
            action_parsed = self.__parse_action(action)
            # Create a java object that can handle multiple java types
            # `ActionSpace` is a class that has been defined in AnyLogic from
            # the ALPypeRLConnector.
            action_space = self.anylogic_model.jvm.com.alpyperl.ActionSpace(action_parsed)
            # Pass action to AnyLogic model.
            self.anylogic_model.step(action_space)

        # Get observation state or sample if in server mode.
        state = (
            np.asarray(list(self.anylogic_model.getState()))
            if not self.server_mode_on
            else self.observation_space.sample()
        )

        # Get 'current' reward (not cumulated) or dummy 0 if in server mode
        # It is assumed that reward will always be an scalar.
        reward = (
            self.anylogic_model.getReward()
            if not self.server_mode_on
            else 0
        )

        # Check if simulation has finished.
        # Simulation length can be fixed or subject to other
        # conditions (e.g. system fails earlier and continuation is non-sense)
        done = (
            self.anylogic_model.hasFinished()
            if not self.server_mode_on
            else True
        )

        # Return tuple: STATE, REWARD, DONE, INFO
        return state, reward, done, False, {}


    def reset(self, *, seed=None, options=None):
        """`[INTERNAL]` Reset function will restart the AnyLogic model to its initial status
        and return the new initial state"""
        # Reset simulation to restart from initial conditions
        new_state = (
            np.asarray(list(self.anylogic_model.reset()))
            if not self.server_mode_on
            else self.observation_space.sample()
        )
        # Return tuble: STATE, INFO
        return new_state, {}


    def render(self):
        """`[INTERNAL]` Whether any visualisation will be displayed or not, depends on the
        user when decides to export an experiment with visualisation or not"""
        pass

    def close(self):
        """`[INTERNAL]` Close executables if any was created"""
        self.anylogic_connector.close_connection()

    def __parse_action(self, action):
        """Parse the action from `numpy` to a primitive type that can be taken by
        java"""
        if isinstance(self.action_space, spaces.Discrete):
            return int(action)
        elif isinstance(self.action_space, spaces.Box) and action.size == 1:
            return float(action[0])

        # Assume it is an array and create a double[] type that can be consumed
        # by java model
        # First get double class from JVM
        double_class = self.anylogic_model.jvm.double
        # Create double array using 'py4j'
        double_array = self.anylogic_model.new_array(double_class, action.size)
        # Populate array with values from action
        for i, v in enumerate(action):
            double_array[i] = float(v)
        return double_array



from alpyperl.examples.cartpole_v1.cartpole_env import CartPoleEnv
import gymnasium as gym
from gymnasium import register


register(
    id='ALCartPole-v1',
    entry_point='alpyperl.examples.cartpole_v1.cartpole_env:CartPoleEnv',
    max_episode_steps=2000,
    disable_env_checker=True
)


import logging
import numpy as np
from py4j.clientserver import ClientServer, JavaParameters, PythonParameters
from threading import Event, Thread
import socket
from alpyperl.anylogic.model.launcher import ALModelLauncher


def get_open_port():
    """This method allows finding free ports in host machine"""
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("",0))
    s.listen(1)
    port = s.getsockname()[1]
    s.close()
    return port


class AnyLogicModelCallback:
    """This is the implementation of the interface that has been defined in the
    Java side and allows AnyLogic to know what functions is allowed to access
    """
    # -----------------------------------------------------------------
    # Code here (any method created here needs to be added to the AL
    # interface) and must be named using the same structure
    # -----------------------------------------------------------------
    def __init__(self):
        self.thread_handler = None

    def finishedModelSetup(self):
        """This function is called from the Java side to unblock python script
        which is waiting for the simulation to finish loading and setting up
        """
        self.thread_handler.set()
        self.thread_handler.clear()
        return True

    def toString(self):
        """Need 'toString()' implementation because AnyLogic calls this
        to visualise the value of variables
        """
        return "Python interface controller for AnyLogic model in java"

    # It is required to define the Java interface that is implemented. Note that is
    # important to properly define the package name as in AnyLogic
    class Java:
        implements = ["com.alpyperl.RLPythonController"]


class AnyLogicModelConnector:
    """This is the main class that creates the gateway between the python script
    and the AnyLogic model. As well as in charge of executing and setting the
    simulation
    """

    def __init__(
        self,
        run_exported_model,
        exported_model_loc,
        show_terminals
    ):
        self.logger = logging.getLogger(__name__)
        # Initialise model launcher
        self.al_model_launcher = None
        # Create an instance of the python implementation to be accessed by the
        # AnyLogic model
        self.anylogic_model_callback = AnyLogicModelCallback()

        # Create single thread gateway between python and AnyLogic (only possible with
        # ClientServer)
        # We want it to be single-thread to force things to happen in a certain
        # order as defined by the python script and not in parallel (asyncronously)
        #
        # E.g. if python script takes some seconds to calculate a decision, the simulation
        # should be waiting for it to finish because it requires the result of the
        # operation to continue (e.g. next action)

        # Dynamic ports is necessary to allow multiple runs to be executed in
        # parallel.
        # This wouldn't be possible using the default configuration because it always
        # uses the same default ports (Java port: 25333, Python port: 25334)

        # This is the port that will be used from the Java side
        java_port = get_open_port() if run_exported_model else 25333

        # Connect python side to Java side with Java dynamic port and start python
        # callback server with a dynamic port (port=0)
        self.gateway = ClientServer(
            java_parameters=JavaParameters(auto_field=True, port=java_port),
            python_parameters=PythonParameters(port=0 if run_exported_model else 25334),
            python_server_entry_point=self.anylogic_model_callback
        )
        # Retrieve the port on which the python callback server was bound to.
        python_port = (
            self.gateway.get_callback_server().get_listening_port()
        )
        self.logger.debug(
            "ALPypeRL connection will take place on ports "
            f"(Java: {java_port} / Python: {python_port})"
        )
        # Execute AnyLogic anylogic_model and tell to which ports it needs to
        # connect. All of this will be handled by the 'ALModelLauncher'
        def execute_model():
            self.al_model_launcher = ALModelLauncher(
                java_port=java_port,
                python_port=python_port,
                folder_location=exported_model_loc,
                show_terminals=show_terminals
            )
            self.al_model_launcher.compile_and_run()
        # Run the model in another thread to avoid this being called from the same thread
        # as the python script, which will cause the execution to be on hold (due to the
        # AL model waiting for instructions - listening)
        if run_exported_model:
            thread = Thread(target=execute_model, args=[])
            thread.start()
        else:
        # In case the user wants to run the model from AnyLogic directly (not 
        # using the exported version) the script will notify him/her that the
        # python script is waiting and ready for the AnyLogic model to be launched
            self.logger.info(
                "You can now launch your AnyLogic model! "
                "'ALPypeRLConnector' will handle the connection for you."
            )
        # The following code will block any further execution of the python code
        # until the AnyLogic model calls back and unblocks the event via the
        # python interface
        thread_handler = Event()
        self.anylogic_model_callback.thread_handler = thread_handler
        self.logger.debug(
            "Python AnyLogic connection handler is waiting AnyLogic model side "
            "to launch and connect"
        )
        thread_handler.wait()

    def close_connection(self):
        """Close model and connection"""
        if self.al_model_launcher is not None:
            self.al_model_launcher.close_model()



from setuptools import setup, find_packages

# Read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(name='alpyperl',
      version='0.0.17',
      license='Apache License, Version 2.0',
      author='Marc Escandell Mari',
      author_email='marcescandellmari@gmail.com',
      description='An open source library for connecting AnyLogic models with Reinforcement Learning frameworks through OpenAI Gymnasium ',
      long_description=long_description,
      long_description_content_type="text/markdown",
      python_requires='>=3.9, <3.11',
      packages=find_packages(),
      url='https://github.com/MarcEscandell/ALPypeRL',
      keywords='alpyperl',
      install_requires=[
            'gymnasium',
            'py4j',
            'ray',
            'ray[rllib]>=2.3.0',
            'tensorflow',
            'torch',
            'uvicorn',
            'fastapi'
      ]
)


import uvicorn
from typing import List
from fastapi import FastAPI, status
from fastapi.responses import HTMLResponse, JSONResponse
from fastapi.encoders import jsonable_encoder
from pydantic import BaseModel


def launch_policy_server(
    policy_config,
    env,
    env_config=None,
    trained_policy_loc='./"trained_policy"',
    host="0.0.0.0",
    port=3000
):
    """Launch server and host trained policy to allow requests. The server
    requires an observation in the form of an array and will return and action
    (type depends on action space)

    :param policy_config: It refers to the policy (also refered as *RL algorithm*) that 
        will be trained. It must be an instance of **rllib algorithms** (check here 
        for more `information <https://docs.ray.io/en/latest/rllib/rllib-algorithms.html>`_)
    :type env_config: ray.rllib.algorithms
    :param env: The environment or **AnyLogic simulation** that the policy will use
        to train against.
    :type env: alpyperl.BaseAnyLogicEnv
    :param env_config: Any option that will be consumed by the environment. Options
        available are: *TODO:*
    :type env_config: dict
    :param trained_policy_loc: The location of the **rllib** trained policy. Be mindful
        that the rllib policy creates an aditional folder when calling ``save``.
        ``checkpoint_dir = policy.save("./resources/trained_policies/cartpole_v0")``
    :type trained_policy_loc: str
    :param host: The host ID to be used. Defaults to ``localhost`` or ``0.0.0.0``
    :type host: str
    :param port: The port the service will connect to. Defaults to ``3000``
    :type port: int
    """

    # Set server flag on to avoid loading the AnyLogic model
    if env_config is None:
        env_config = {}
    env_config['server_mode_on'] = True

    # Re-create policy configuration with no workers and avoid launching
    # unnecessary models
    policy = (
        policy_config
            .rollouts(num_rollout_workers=0)
            .environment(env=env, env_config=env_config)
            .build()
    )

    # Restore policy state from given checkpoint
    policy.restore(trained_policy_loc)

    # Initialise FastAPI application server
    app = FastAPI()

    @app.get("/")
    async def greetings():
        html_content = """
            <p style=\"text-align:center;font-size:35px\">
                Welcome to <b>ALPypeRL</b> trained policy serving!
            </p>
            <p style=\"text-align:center;font-size:15px\">
                You have accessed the API server. Append <b>'/docs'</b> at the 
                end of this url to access the documentation on the methods available.
            </p>
        """
        return HTMLResponse(content=html_content, status_code=200)

    @app.post("/predict")
    async def predict_next_action(observation: List[float]):
        # Check documentation at https://docs.ray.io/en/latest/serve/tutorials/rllib.html
        action = policy.compute_single_action(observation)
        # Format response
        response = {
            "observation": observation,
            "action": action.tolist()
        }

        return JSONResponse(content=jsonable_encoder(response), status_code=200)

    # Launch server
    uvicorn.run(app, host=host, port=port)


import os
import platform
import re
import string
import random
import subprocess
from subprocess import Popen
from pathlib import Path
import logging


class ALModelLauncher():
    """This class in in charge of launching the AnyLogic model. The main
    functionality comes by notifying the model what ports need to be connecting
    to so communication with python script is possible.
    """

    def __init__(
        self,
        java_port=25333,
        python_port=25334,
        folder_location='./exported_model',
        show_terminals=False
    ):
        self.logger = logging.getLogger(__name__)
        # Check if exported model exists.
        if not os.path.exists(folder_location):
            raise Exception(
                f"Could not find exported model folder '{folder_location}'. "
                "Model execution could not proceed. Please export your model at "
                f"'{folder_location}'"
            )
        # Get operating system compiling this code
        self.os_name = platform.system()
        self.folder_location = Path(folder_location)
        self.show_terminals = show_terminals
        self.java_exec_end_pattern = '%' if self.os_name == 'Windows' else '$'
        # Initialise executable file location and process
        self.exec_loc = None
        self.al_process = None
    
        # Create command-line arguments that refer to the Java and Python ports
        # java model needs to connect to.
        self.port_arg_str = '-jp ' + str(java_port) + ' -pp ' + str(python_port)
        # Get project name.
        self.project_name = self.__get_project_name(folder_location=folder_location)

    def compile_and_run(self):
        """Compile the script and execute it considering the operating system
        it is running on"""

        # Execute exported version of the model depending on OS
        if self.os_name == 'Linux':
            # Generate executable file
            self.exec_loc = f"./{self.__create_exec_file('linux.sh')}"
            # Execute model
            self.al_process = (
                Popen(['gnome-terminal', '--', self.exec_loc]) 
                if self.show_terminals 
                else Popen(['bash', '-c', self.exec_loc])
            )

        elif self.os_name == 'Windows':
            # Generate executable file
            self.exec_loc = self.__create_exec_file('windows.bat')
            # Execute model
            self.al_process = (
                Popen(['start', 'cmd.exe', '/c', self.exec_loc], shell=True)
                if self.show_terminals 
                else Popen([self.exec_loc])
            )

        elif self.os_name == 'Darwin':
            # Generate executable file
            self.exec_loc = self.__create_exec_file('mac')
            # Execute model
            self.al_process = (
                Popen(['open', self.exec_loc])
                if self.show_terminals 
                else Popen([f'./{self.exec_loc}'])
            )
        self.logger.debug(
            f"AnyLogic model '{self.project_name}' has been succesfully "
            "compiled and launched"
        )

    def close_model(self):
        """ Delete model executable file and close process"""
        os.remove(self.exec_loc)
        self.al_process.terminate()
        self.logger.debug("AnyLogic models have been terminated")

    def __get_project_name(self, folder_location):
        """ A function to extract the AnyLogic project name automatically based on
        the expected files to be found after exporting the model to a standalone
        """
        for f in os.listdir(folder_location):
            matches = re.search('(.*)_linux\.sh', f)
            if matches:
                return matches.group(1)

        raise Exception(
            f"Could not find any executable file in '{folder_location}'. "
            "Have you exported your model correctly?"
        )

    def __id_generator(self, size=6, chars=string.ascii_uppercase + string.digits):
        """This is to avoid re-writting the an active script. It can happen when
        asyncronous execution"""
        return ''.join(random.choice(chars) for _ in range(size))

    def __create_exec_file(self, file_type):
        """Insert ports information and create executable file"""
        # Get file extension
        file_extension = Path(file_type).suffix
        # Read file as string
        with open(self.folder_location.joinpath(f'{self.project_name}_{file_type}'), 'r') as file:
            orig_exec_file_str = file.read()
        # Apply regex to insert ports information
        exec_file_str = re.sub(
            f'([\s\S]*)(\{self.java_exec_end_pattern}\*)([\s\S]*)',
            f'\g<1>{self.port_arg_str} \g<2>\g<3>',
            orig_exec_file_str
        )
        # File name and location
        exec_loc = self.folder_location.joinpath(f'{self.project_name}-{self.__id_generator()}{file_extension}')
        # Create and write bash script
        with open(exec_loc, "w+") as file:
            file.write(exec_file_str)
        # Parse text file to unix executable extension only if in Linux system
        if not self.os_name == 'Windows':
            os.system(f'chmod +x {exec_loc}')

        return exec_loc
